/*
 * Copyright 2019 Dan Maas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package io.github.resilience4j.common.ratelimiter.configuration;

import java.time.Duration;
import java.util.HashMap;
import java.util.Map;

import javax.validation.constraints.Min;

import io.github.resilience4j.common.utils.ConfigUtils;
import io.github.resilience4j.core.ConfigurationNotFoundException;
import io.github.resilience4j.core.StringUtils;
import io.github.resilience4j.core.lang.Nullable;
import io.github.resilience4j.ratelimiter.RateLimiterConfig;

public class RateLimiterConfigurationProperties {

	private Map<String, InstanceProperties> instances = new HashMap<>();
	private Map<String, InstanceProperties> configs = new HashMap<>();

	public RateLimiterConfig createRateLimiterConfig(@Nullable InstanceProperties instanceProperties) {
		if (instanceProperties == null) {
			return RateLimiterConfig.ofDefaults();
		}
		if (StringUtils.isNotEmpty(instanceProperties.getBaseConfig())) {
			InstanceProperties baseProperties = configs.get(instanceProperties.baseConfig);
			if (baseProperties == null) {
				throw new ConfigurationNotFoundException(instanceProperties.getBaseConfig());
			}
			return buildConfigFromBaseConfig(baseProperties, instanceProperties);
		}
		return buildRateLimiterConfig(RateLimiterConfig.custom(), instanceProperties);
	}

	private RateLimiterConfig buildConfigFromBaseConfig(InstanceProperties baseProperties, InstanceProperties instanceProperties) {
		ConfigUtils.mergePropertiesIfAny(baseProperties, instanceProperties);
		RateLimiterConfig baseConfig = buildRateLimiterConfig(RateLimiterConfig.custom(), baseProperties);
		return buildRateLimiterConfig(RateLimiterConfig.from(baseConfig), instanceProperties);
	}

	private RateLimiterConfig buildRateLimiterConfig(RateLimiterConfig.Builder builder, @Nullable InstanceProperties instanceProperties) {
		if (instanceProperties == null) {
			return builder.build();
		}

		if (instanceProperties.getLimitForPeriod() != null) {
			builder.limitForPeriod(instanceProperties.getLimitForPeriod());
		}

		if (instanceProperties.getLimitRefreshPeriod() != null) {
			builder.limitRefreshPeriod(instanceProperties.getLimitRefreshPeriod());
		}

		if (instanceProperties.getTimeoutDuration() != null) {
			builder.timeoutDuration(instanceProperties.getTimeoutDuration());
		}

		return builder.build();
	}

	private InstanceProperties getLimiterProperties(String limiter) {
		return instances.get(limiter);
	}

	public RateLimiterConfig createRateLimiterConfig(String limiter) {
		return createRateLimiterConfig(getLimiterProperties(limiter));
	}

	@Nullable
	public InstanceProperties getInstanceProperties(String instance) {
		return instances.get(instance);
	}

	public Map<String, InstanceProperties> getInstances() {
		return instances;
	}

	/**
	 * For backwards compatibility when setting limiters in configuration properties.
	 */
	public Map<String, InstanceProperties> getLimiters() {
		return instances;
	}

	public Map<String, InstanceProperties> getConfigs() {
		return configs;
	}

	/**
	 * Class storing property values for configuring {@link RateLimiterConfig} instances.
	 */
	public static class InstanceProperties {

		private Integer limitForPeriod;
		private Duration limitRefreshPeriod;
		private Duration timeoutDuration;
		@Nullable
		private Boolean subscribeForEvents;
		@Nullable
		private Boolean registerHealthIndicator;
		@Min(1)
		@Nullable
		private Integer eventConsumerBufferSize;
		@Nullable
		private String baseConfig;

		/**
		 * Configures the permissions limit for refresh period.
		 * Count of permissions available during one rate limiter period
		 * specified by {@link RateLimiterConfig#getLimitRefreshPeriod()} value.
		 * Default value is 50.
		 *
		 * @return the permissions limit for refresh period
		 */
		@Nullable
		public Integer getLimitForPeriod() {
			return limitForPeriod;
		}

		/**
		 * Configures the permissions limit for refresh period.
		 * Count of permissions available during one rate limiter period
		 * specified by {@link RateLimiterConfig#getLimitRefreshPeriod()} value.
		 * Default value is 50.
		 *
		 * @param limitForPeriod the permissions limit for refresh period
		 */
		public InstanceProperties setLimitForPeriod(Integer limitForPeriod) {
			this.limitForPeriod = limitForPeriod;
			return this;
		}

		/**
		 * Configures the period of limit refresh.
		 * After each period rate limiter sets its permissions
		 * count to {@link RateLimiterConfig#getLimitForPeriod()} value.
		 * Default value is 500 nanoseconds.
		 * @deprecated As of release 0.16.0 , use {@link #getLimitRefreshPeriod()} instead
		 * @return the period of limit refresh
		 */
		@Deprecated
		@Nullable
		public Integer getLimitRefreshPeriodInNanos() {
			if (limitRefreshPeriod != null) {
				return (int) limitRefreshPeriod.toNanos();
			} else {
				return null;
			}
		}

		/**
		 * Configures the period of limit refresh.
		 * After each period rate limiter sets its permissions
		 * count to {@link RateLimiterConfig#getLimitForPeriod()} value.
		 * Default value is 500 nanoseconds.
		 * @deprecated As of release 0.16.0 , use {@link #setLimitRefreshPeriod(Duration)} instead
		 * @param limitRefreshPeriodInNanos the period of limit refresh
		 */
		@Deprecated
		public InstanceProperties setLimitRefreshPeriodInNanos(Integer limitRefreshPeriodInNanos) {
			this.limitRefreshPeriod = Duration.ofNanos(limitRefreshPeriodInNanos);
			return this;
		}

		/**
		 * Configures the period of limit refresh.
		 * After each period rate limiter sets its permissions
		 * count to {@link RateLimiterConfig#getLimitForPeriod()} value.
		 * Default value is 500 nanoseconds.
		 *
		 * @return the period of limit refresh
		 */
		@Nullable
		public Duration getLimitRefreshPeriod() {
			return limitRefreshPeriod;
		}

		/**
		 * Configures the period of limit refresh.
		 * After each period rate limiter sets its permissions
		 * count to {@link RateLimiterConfig#getLimitForPeriod()} value.
		 * Default value is 500 nanoseconds.
		 *
		 * @param limitRefreshPeriod the period of limit refresh
		 */
		public InstanceProperties setLimitRefreshPeriod(Duration limitRefreshPeriod) {
			this.limitRefreshPeriod = limitRefreshPeriod;
			return this;
		}

		/**
		 * Configures the default wait for permission duration.
		 * Default value is 5 seconds.
		 * @deprecated As of release 0.16.0 , use {@link #getTimeoutDuration()}instead
		 * @return wait for permission duration
		 */
		@Deprecated
		@Nullable
		public Integer getTimeoutInMillis() {
			if (timeoutDuration != null) {
				return (int) timeoutDuration.toMillis();
			} else {
				return null;
			}
		}

		/**
		 * Configures the default wait for permission duration.
		 * Default value is 5 seconds.
		 * @deprecated As of release 0.16.0 , use {@link #setTimeoutDuration(Duration)} instead
		 * @param timeoutInMillis wait for permission duration
		 */
		@Deprecated
		public InstanceProperties setTimeoutInMillis(Integer timeoutInMillis) {
			this.timeoutDuration = Duration.ofMillis(timeoutInMillis);
			return this;
		}

		/**
		 * Configures the default wait for permission duration.
		 * Default value is 5 seconds.
		 *
		 * @return wait for permission duration
		 */
		@Nullable
		public Duration getTimeoutDuration() {
			return timeoutDuration;
		}

		/**
		 * Configures the default wait for permission duration.
		 * Default value is 5 seconds.
		 *
		 * @param timeout wait for permission duration
		 */
		public InstanceProperties setTimeoutDuration(Duration timeout) {
			this.timeoutDuration = timeout;
			return this;
		}

		public Boolean getSubscribeForEvents() {
			return subscribeForEvents;
		}

		public InstanceProperties setSubscribeForEvents(Boolean subscribeForEvents) {
			this.subscribeForEvents = subscribeForEvents;
			return this;
		}

		public Integer getEventConsumerBufferSize() {
			return eventConsumerBufferSize;
		}

		public InstanceProperties setEventConsumerBufferSize(Integer eventConsumerBufferSize) {
			this.eventConsumerBufferSize = eventConsumerBufferSize;
			return this;
		}

		public Boolean getRegisterHealthIndicator() {
			return registerHealthIndicator;
		}

		public InstanceProperties setRegisterHealthIndicator(Boolean registerHealthIndicator) {
			this.registerHealthIndicator = registerHealthIndicator;
			return this;
		}

		/**
		 * Gets the shared configuration name. If this is set, the configuration builder will use the the shared
		 * configuration instance over this one.
		 *
		 * @return The shared configuration name.
		 */
		@Nullable
		public String getBaseConfig() {
			return baseConfig;
		}

		/**
		 * Sets the shared configuration name. If this is set, the configuration builder will use the the shared
		 * configuration instance over this one.
		 *
		 * @param baseConfig The shared configuration name.
		 */
		public InstanceProperties setBaseConfig(String baseConfig) {
			this.baseConfig = baseConfig;
			return this;
		}
	}

}
